/*
 * Copyright (C) 1997  Warren Jasper
 * All rights reserved.
 *
 * This software may be freely copied, modified, and redistributed
 * provided that this copyright notice is preserved on all copies.
 *
 * You may not distribute this software, in whole or in part, as part of
 * any commercial product without the express consent of the authors.
 *
 * There is no warranty or other guarantee of fitness of this software
 * for any purpose.  It is provided solely "as is".
 *
 *
 *   Note: Interrupts, when enabled, are generated differently depending
 *   upon the mode of the board:
 *
 *   Soft Trigger:   poll IRQDATA for A/D done.
 *   Pacer clock:    interrupts generated with FIFO half full
 *                   or when Total_Count = 0.
 */

/***************************************************************************
 *
 * a2dc.c
 *
 ***************************************************************************/

#define MODULE

#include <linux/version.h>
#include <linux/module.h>
#include <linux/types.h>
#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/major.h>
#include <linux/mm.h>
#include <linux/timer.h>
#include <linux/ioport.h>
#include <linux/kernel.h>
#include <linux/malloc.h>
#include <linux/string.h>
#include <linux/signal.h>
#include <asm/io.h>
#include <asm/segment.h>
#include <asm/system.h>
#include <stdlib.h>
#include "adc.h"
#include "a2dc.h"

/***************************************************************************
 *
 * Prototype of pubic and private functions.
 *
 ***************************************************************************/

static int  SoftRead(WORD *outBuff, ADC_ChanRec *chan);
static int  PacerRead(WORD *outBuff, ADC_ChanRec *chan);
static int  ExternRead(WORD *outBuff, ADC_ChanRec *chan);
static void SetGainQueue( BYTE addr, BYTE chan, BYTE gain);
static int  SetPacerFreq(ADC_BoardRec *board);
static void StopPacer();
static void LoadPacer(ADC_BoardRec *board);
static void LoadCounter0( LONG arg );
static void LoadTotalCount( LONG totalCount );
static void ResetFIFO();

static int adc_read(struct inode *iNode, struct file *filePtr, char *buf, int count);
static int dio_write(struct inode *iNode, struct file *filePtr, const char *buf, int count);
static int adc_open(struct inode *iNode, struct file *filePtr);
static void adc_close(struct inode *iNode, struct file *filePtr);
static int adc_ioctl(struct inode *iNode, struct file *filePtr, unsigned int cmd, LONG arg);
static void adc0_ReadInterrupt();
static void adc0_TimerHandler(LONG);
static int adc_find( WORD base_reg );
int init_module(void);
void cleanup_module(void) ;

/***************************************************************************
 *
 * Global data.
 *
 ***************************************************************************/

static WORD base_reg =  PORT_MIN;           /* base register address */
static WORD base_reg2;                      /* base register +0x400  */
static int WordsToRead;                     /* number of conversions until done  */
static BYTE dio_reg = 0x80;                 /* DIO control register              */
static ADC_ChanRec *CurrChan = NULL;        /* pointer to Chan[minor]            */
static int MajorNumber = DEFAULT_MAJOR_DEV; /* Major number compiled in          */
static ADC_ChanRec Chan[AD_CHANNELS];       /* ADC Channel specific information  */
static DIO_ChanRec ChanDIO[DIO_PORTS];      /* DIO Channel specific information  */
static ADC_BoardRec BoardData;              /* Board specific information        */
static WORD KernBuff[MAX_COUNT];            /* Kernel buffer where samples are   */
                                            /* saved before write to user space  */
static WORD* KernBuffPtr;                   /* pointer to kernbuf                */
#if LINUX_VERSION_CODE < 0x020000
char kernel_version[] = UTS_RELEASE;        /* Required to by module             */
#endif
struct wait_queue  *adc0_wait;              /* wait semaphore                    */
struct timer_list TimerList = {NULL, NULL, 0, 0, adc0_TimerHandler};

/***************************************************************************
 *
 *
 ***************************************************************************/

static struct file_operations adc_fops = {
	NULL,           /* adc_seek */
	adc_read,
	dio_write,      /* dio_write   */
	NULL,		/* adc_readdir */
	NULL,		/* adc_select  */
	adc_ioctl,	/* adc_ioctl   */
	NULL,
	adc_open,
	adc_close,
	NULL		/* fsync       */
};

#ifdef __cplusplus
extern "C" {
#endif

/***************************************************************************
 *
 * Loads driver. Called when "insmod das16-M1.o" is invoked on the command line.
 *               The board is set to IRQ, 
 *
 ***************************************************************************/

int init_module(void) 
{
  int tryings[] = {12,11,10,7,5,3,2,-1};
  int i, err;

  /* Register as a device with kernel.  */

  if ((err = register_chrdev(MajorNumber, "das16/M1", &adc_fops))) {
      printk("%s: Failure to load module. error %d\n", ADAPTER_ID, -err);
      return err;
  } 

  /* Register the region of iospace with kernel and get base address */

#ifdef ADC_BASE
    base_reg = ADC_BASE;
    base_reg2 = base_reg + 0x400;
    if ( check_region(base_reg, PORT1_SIZE ) != 0 ) {  /* in use */
        printk("%s: Can't allocate region port address %#x\n", 
                ADAPTER_ID, base_reg );
        if (unregister_chrdev(MajorNumber, "das16/M1") != 0) {
          printk("%s: unregister_chrdev() failed.\n", ADAPTER_ID);
	}
        return -ENODEV;
    } else  if ( check_region(base_reg2, PORT2_SIZE ) != 0 ) {  /* in use */
        printk("%s: Can't allocate region port address %#x\n", 
                ADAPTER_ID, base_reg );
        if (unregister_chrdev(MajorNumber, "das16/M1") != 0) {
          printk("%s: unregister_chrdev() failed.\n", ADAPTER_ID);
	}
        return -ENODEV;
    } else {
      request_region( base_reg,  PORT1_SIZE, "das16/M1" );
      request_region( base_reg2, PORT2_SIZE, "das16/M1" );
    }
    BoardData.base = base_reg;
    BoardData.base2 = base_reg2;
#else
  /* Look to see if ADC is installed */
  for ( base_reg = PORT_MIN; base_reg <= PORT_MAX; base_reg += PORT_STEP ) {
      base_reg2 = base_reg + 0x400;
      if ( check_region(base_reg,  PORT1_SIZE ) == 0 &&
           check_region(base_reg2, PORT2_SIZE ) == 0    ) {  /* not in use */
          request_region( base_reg,  PORT1_SIZE, "das16/M1" );
          request_region( base_reg2, PORT2_SIZE, "das16/M1" );
          if (adc_find( base_reg )) {  /* found the board */
              BoardData.base = base_reg;
              BoardData.base2 = base_reg2 ;
              break;
          } else {
              release_region( base_reg,  PORT1_SIZE);
              release_region( base_reg2, PORT2_SIZE);
          }
      }
  }

  if ( base_reg > PORT_MAX ) {
      printk("%s: No boards found from address %#x to %#x.\n", 
             ADAPTER_ID, PORT_MIN, PORT_MAX);
      if (unregister_chrdev(MajorNumber, "das16/M1") != 0) {
          printk("%s: unregister_chrdev() failed.\n", ADAPTER_ID);
      }
      return -ENODEV;
  }
#endif

  /* Register interrupt handler. */

#ifdef ADC_IRQ
  BoardData.irq = ADC_IRQ;
#else
  BoardData.irq = -1;

  for ( i = 0; tryings[i] > 0; i++ ) {
#if LINUX_VERSION_CODE > 0x020000
      if (request_irq(tryings[i], adc0_ReadInterrupt, SA_INTERRUPT, "das16/M1", NULL) == 0) {
          free_irq(tryings[i],NULL);
#else 
      if (request_irq(tryings[i], adc0_ReadInterrupt, SA_INTERRUPT, "das16/M1") == 0) { 
          free_irq(tryings[i]);
#endif
          BoardData.irq = tryings[i];
          break;
      }
  }
#endif

  if ( BoardData.irq == -1 ) {
      /* No free irq found! cleanup and exit */
      printk("%s: found board but no free irq.\n", ADAPTER_ID);
      if (unregister_chrdev(MajorNumber, "das16/M1") != 0) {  /* unregister device */
          printk("%s: unregister_chrdev() failed.\n", ADAPTER_ID);
      }
      release_region( BoardData.base,  PORT1_SIZE );       /* release region */
      release_region( BoardData.base2, PORT2_SIZE );       /* release region */
      return -EADDRNOTAVAIL;
  }
#if LINUX_VERSION_CODE > 0x020000
  if (request_irq(BoardData.irq, adc0_ReadInterrupt, SA_INTERRUPT, "das16/M1", NULL)) {
#else
  if (request_irq(BoardData.irq, adc0_ReadInterrupt, SA_INTERRUPT, "das16/M1")) {
#endif
    printk("%s: Can't request IRQ %d\n", ADAPTER_ID, BoardData.irq);
    if ( unregister_chrdev( MajorNumber, "das16/M1" ) != 0 ) {  /* unregister device */
        printk( "%s: unregister_chrdev() failed.\n", ADAPTER_ID);
    }
    release_region( BoardData.base,  PORT1_SIZE );       /* release region */
    release_region( BoardData.base2, PORT2_SIZE );       /* release region */
    return -(EIO);
  }

  printk("%s: address=%#x IRQ=%d.",ADAPTER_ID, BoardData.base, BoardData.irq);
  printk(" 10/3/97 wjasper@tx.ncsu.edu\n");

  /* Set all channel structures to show nothing active/open */
  for (i = 0; i < AD_CHANNELS; i++) {
      Chan[i].open = FALSE;
      Chan[i].adc_dio_pretrig = FALSE;
      Chan[i].chan = i;
      Chan[i].update_queue = FALSE;
  }

  ChanDIO[0].addr = DIO_PORTA;
  ChanDIO[1].addr = DIO_PORTB;
  ChanDIO[2].addr = DIO_PORTC;

  for ( i = 0; i < DIO_PORTS; i++ ) {
    ChanDIO[i].open = FALSE; 
    ChanDIO[i].mode = 0;
    outb_p(0x0, ChanDIO[i].addr);
  }

  adc0_wait = NULL;

  /* 
     Set interrupt level on board  
     IRQ        Mask
     ---        ----
     2     ->   0x40
     3     ->   0x50
     5     ->   0x60
     7     ->   0x70
     10    ->   0x00
     11    ->   0x10
     12    ->   0x20
     15    ->   0x30
  */

  switch( BoardData.irq ) {
    case 2:
      BoardData.ir = 0x40;
      break;
    case 3:
      BoardData.ir = 0x50;
      break;
    case 5:
      BoardData.ir = 0x60;
      break;
    case 7:
      BoardData.ir = 0x70;
      break;
    case 10:
      BoardData.ir = 0x00;
      break;
    case 11:
      BoardData.ir = 0x10;
      break;
    case 12:
      BoardData.ir = 0x20;
       break;
    case 15:
      BoardData.ir = 0x30;
      break;
  }

  cli();
  outb_p(BoardData.ir, CNTRL_REG);  /* ... clearing INTE bit in control reg */
  outb_p(0x00, CLR_INTERRUPT);                  /* Clear IRQDATA bit  */
  sti();

  if ( DT_CONNECT ) {
    BoardData.statusReg = DTEN;
  } else {
    BoardData.statusReg = 0x0;
  }
  outb(BoardData.statusReg, STATUS_REG);

  BoardData.freq = DEFAULT_FREQ;   /* Set default pacer clock frequency */
  SetPacerFreq(&BoardData);

  BoardData.busy = FALSE;          /* board ready */

  #ifdef DEBUG
      printk("%s: IRQ set to %d\n", ADAPTER_ID, BoardData.irq);
  #endif

  return 0;
}

/***************************************************************************
 *
 * Test memory locations for exsistance of adc board.
 *
 ***************************************************************************/
static int adc_find( WORD base_reg)
{

/*
    To test if there is a DAS 16/M1 board by checking the dio at
    base_reg + 0x400
*/
 
  BYTE bReg, value;
  WORD i;

  base_reg += 0x400;
  value = 0xa5;

  /* Test the 82C55 */
  dio_reg = 0x80;
  outb_p( dio_reg, base_reg+3);      /* put in output mode 0 ? */

  for ( i = 0; i < DIO_PORTS; i++ ) {
    outb_p(value, base_reg + i);
    bReg = inb_p(base_reg + i);

    if ( bReg != value ) {
        return FALSE;   /* not the DAS16-M1 */
    }
  }

  return TRUE;                       /* found it! */
}

/***************************************************************************
 *
 * Remove driver. Called when "rmmod das16/M1" is run on the command line.
 *
 ***************************************************************************/

void cleanup_module(void) 
{
  if (MOD_IN_USE) {
    printk("%s: device busy, remove delayed.\n", ADAPTER_ID);
    return;
  }

  StopPacer();

  release_region( BoardData.base,  PORT1_SIZE );
  release_region( BoardData.base2, PORT2_SIZE );

#if LINUX_VERSION_CODE > 0x020000
  free_irq(BoardData.irq, NULL);
#else
  free_irq(BoardData.irq);
#endif
  if (unregister_chrdev(MajorNumber, "das16/M1") != 0) {
    printk("%s: cleanup_module failed.\n", ADAPTER_ID);
  } else {

  #ifdef DEBUG
      printk("%s: module removed.\n", ADAPTER_ID);
  #endif

  }    
}

#ifdef __cplusplus
}
#endif

/***************************************************************************
 *
 * open() service handler
 *
 ***************************************************************************/

static int adc_open(struct inode *iNode, struct file *filePtr)
{
  int minor = MINOR(iNode->i_rdev);
  int port;

  MOD_INC_USE_COUNT;

  /* 
     check if device is already open: only one process may read from a
     port at a time.  There is still the possibility of two processes 
     reading from two different channels messing things up. However,
     the overhead to check for this may not be worth it.
  */

  if ( minor >= 0  && minor < AD_CHANNELS ) {
    if ( Chan[minor].open == TRUE ) {
      return -EBUSY;
    }

    Chan[minor].open = TRUE;              /* The device is open */
    Chan[minor].gain = BP_10_00V;         /* +/- 10V */
    Chan[minor].mode = filePtr->f_flags;  /* set acquisition mode */

    outb_p(0x00, CLR_INTERRUPT);          /* Clear IRQDATA bit  */

    #ifdef DEBUG
        printk("%s: open(): minor %d mode %d.\n", ADAPTER_ID, minor, Chan[minor].mode);
    #endif
    return 0;   
  }

  if ( minor >= AD_CHANNELS  && minor < AD_CHANNELS + DIO_PORTS ) {
    port = minor - AD_CHANNELS;
    if ( ChanDIO[port].open == TRUE ) {
      return -EBUSY;
    }

    ChanDIO[port].open = TRUE;                 /* The device is open */
    ChanDIO[port].f_flags = filePtr->f_flags;

    #ifdef DEBUG
      printk("%s: open(): minor %d mode %d.\n", ADAPTER_ID, minor, ChanDIO[port].mode);
    #endif
    return 0;
  }

  return 0;
}

/***************************************************************************
 *
 * close() service handler
 *
 ***************************************************************************/

static void adc_close(struct inode *iNode, struct file *filePtr)
{
  int minor = MINOR(iNode->i_rdev);
  int port;

  #ifdef DEBUG
      printk("%s: close() of minor number %d.\n", ADAPTER_ID, minor);
  #endif

  MOD_DEC_USE_COUNT;

  if ( minor >= 0  && minor < AD_CHANNELS ) {
    Chan[minor].open = FALSE;
  }

  if ( minor >= AD_CHANNELS  && minor < AD_CHANNELS + DIO_PORTS ) {
    port = minor - AD_CHANNELS;
    ChanDIO[port].open = FALSE;
  }
}

/***************************************************************************
 *
 * read() service function
 *
 ***************************************************************************/

static int adc_read(struct inode *iNode, struct file *filePtr, char *buf, int count)
{
  int   stat;
  BYTE  bReg;
  int   minor = MINOR(iNode->i_rdev);
  int   port;
  register int i;

  if ( BoardData.busy == TRUE ) {   /* if board is in use, return busy */
      return (-EBUSY);
  } else {
      BoardData.busy = TRUE;
  }

  if ( count < 1 ) {
    printk("adc_read(): count must be greater than 0.\n");
    return (-1);
  }

  if ( minor >= 0 && minor < AD_CHANNELS ) {

    /* Set modes for this channel */
    if ( !Chan[minor].update_queue ) {
      SetGainQueue((BYTE) 0, Chan[minor].chan, Chan[minor].gain);
    }

    if ( (Chan[minor].count = count) > MAX_COUNT ) {
        printk("adc_read(): requesting too large a count size.\n");
        BoardData.busy = FALSE;
        return (-1);
    }

    /* Read */

    switch (Chan[minor].mode) {

      case ADC_SOFT_TRIGGER:
          #ifdef DEBUG
              printk("adc_read(): Entering ADC_SOFT_TRIGGER mode.\n");
          #endif

          if (SoftRead(KernBuff, &Chan[minor])) {
              printk("adc_read: SoftTrigRead() failed.\n");
              BoardData.busy = FALSE;
              return(-1);
          } 
          break;

      case ADC_EXTERNAL_PACER:

          #ifdef DEBUG
              printk("adc_read(): Entering ADC_EXTERNAL_PACER mode.\n");
          #endif

          BoardData.statusReg = inb(STATUS_REG) & ~TRIG0 ;
          outb(BoardData.statusReg, STATUS_REG);
          bReg = inb(CNTRL_REG) & ~(INTEN | S0);
          bReg |= S1;
          outb(bReg, CNTRL_REG);
          outb(0x00, CLR_INTERRUPT);                  /* Clear IRQDATA bit  */

          if (ExternRead(KernBuff, &Chan[minor])) {
              printk("adc_read: ExternalRead() failed with ADC_EXTERNAL_PACER.\n");
              BoardData.busy = FALSE;
              return(-1);
          }
          break;

      case ADC_PACER_CLOCK:

          #ifdef DEBUG
              printk("adc_read(): Entering ADC_PACER_CLOCK mode.\n");
          #endif

          if ( Chan[minor].count == 1 ) {            /* use SoftTrigger */
              if (SoftRead(KernBuff, &Chan[minor])) {
                  printk("adc_read: SoftRead() failed with pacer.\n");
                  BoardData.busy = FALSE;
                  return(-1);
              }
          } else {
              StopPacer();                        /* disable pacer if in pacer mode */
              BoardData.statusReg = inb(STATUS_REG) & ~TRIG0; 
              outb(BoardData.statusReg, STATUS_REG);
              bReg = inb(CNTRL_REG) & ~(INTEN);
              bReg |= (S1 | S0);
              outb(bReg, CNTRL_REG);
              outb(0x00, CLR_INTERRUPT);         /* Clear IRQDATA bit  */

              if (PacerRead(KernBuff, &Chan[minor])) {
                  printk("adc_read: PacerRead() failed.\n");
                  BoardData.busy = FALSE;
                  return(-1);
              } 
          }
          break;

      case ADC_PACER_EXTERN_TRIG:
          #ifdef DEBUG
              printk("adc_read(): Entering ADC_PACER_EXTERN_TRIG mode.\n");
          #endif

          StopPacer();                           /* disable pacer if in pacer mode */
          BoardData.statusReg = inb(STATUS_REG) | TRIG0; 
          outb(BoardData.statusReg, STATUS_REG);
          bReg = inb(CNTRL_REG) & ~(INTEN);
          bReg |= (S1 | S0);
          outb(bReg, CNTRL_REG);
          outb(0x00, CLR_INTERRUPT);          /* Clear IRQDATA bit  */

          if (PacerRead(KernBuff, &Chan[minor])) {
              printk("adc_read: ExternRead() failed with ADC_PACER_EXTERN_TRIG.\n");
              BoardData.busy = FALSE;
              return(-1);
          } 
          break;
    }

    /* Check for overflows */
    if ( (bReg = (inb_p(STATUS_REG) & OVRUN)) ) {
      printk("adc_read: FIFO Buffer overflow or DT-Connect handshake failed.\n");
      StopPacer();
      outb(bReg & ~(DTEN), STATUS_REG);    /* clear OVRUN bit */
      BoardData.busy = FALSE;
      return -1;
    }

    /* Check that data can be written to file. */

    if ((stat = verify_area(VERIFY_WRITE, buf, sizeof(WORD)*Chan[minor].count)) != 0) {
      printk("adc_read: Failed VERIFY_WRITE.\n");
      BoardData.busy = FALSE;
      return -1;
    }

    /* Write data to user space */
    if ( !Chan[minor].update_queue ) {
      for (i = 0; i < Chan[minor].count; i++) {
        KernBuff[i] >>= 4;
      }
    }

    if (Chan[minor].count == 1) {
      put_fs_word(*KernBuff, (WORD*) buf);
    } else  {
      memcpy_tofs(buf, KernBuff, Chan[minor].count*sizeof(WORD));
    }

    BoardData.busy = FALSE;
    return(Chan[minor].count);     /* return number of samples read */
  }

  /* check to see if reading a value from the DIO */

  if ( minor >= AD_CHANNELS  && minor < AD_CHANNELS + DIO_PORTS ) {
    port = minor - AD_CHANNELS;
    bReg = inb(ChanDIO[port].addr);
    put_fs_byte(bReg, buf);

    #ifdef DEBUG
       printk("adc_read: %s DIO Port %#x set to %#x\n", ADAPTER_ID, minor, bReg);
    #endif
    return 1;
  }

  printk("adc_read: Incorrect minor number (%d).\n", minor);
  BoardData.busy = FALSE;
  return -1;
}

/***************************************************************************
 *
 * write() service function
 *
 ***************************************************************************/

static int dio_write(struct inode *iNode, struct file *filePtr, const char *buf, int count)
{
    int  minor = MINOR(iNode->i_rdev);
    int  port;
    BYTE bReg;

  /* check to see if writing a value to the DIO */
  if ( minor >= AD_CHANNELS  && minor < AD_CHANNELS + DIO_PORTS ) {
    port = minor - AD_CHANNELS;
    bReg = get_fs_byte(buf);
    ChanDIO[port].value = bReg;
    outb( bReg, ChanDIO[port].addr);
  } else {
    printk("adc_write: Incorrect minor number (%d).\n", minor);
    return -1;
  }
  #ifdef DEBUG
    printk("dio_write %s: DIO Port %#x set to %#x\n", ADAPTER_ID, minor, bReg);
  #endif
  return 1;
}

/***************************************************************************
 *
 * iotctl() service handler
 *
 ***************************************************************************/

/* Note:
    Remember that FIOCLEX, FIONCLEX, FIONBIO, and FIOASYN are reserved ioctl cmd numbers
*/

static int adc_ioctl(struct inode *iNode, struct file *filePtr, unsigned int cmd, LONG arg)
{
  int minor = MINOR(iNode->i_rdev);
  int port;
  BYTE gain, chan, addr, bReg;

  if ( minor >= 0 && minor < AD_CHANNELS ) {
    switch (cmd) {
      case ADC_SET_GAINS:
        if ( Chan[minor].update_queue ) {
           gain = (BYTE) (arg & 0x0000ff);
           chan = (BYTE) ((arg & 0x00ff00) >> 8);
           addr = (BYTE) ((arg & 0xff0000) >> 16);
           if ( addr > 255 ) {
             printk("ADC_SET_GAINS: Gain queue address exceeds 255. addr = %#x.\n", addr);
             return -EINVAL;
           }
           if ( ( addr%2 != chan%2 ) && (addr > 0) ) {
            printk("ADC_SET_GAINS: Even/Odd channels at even/odd address. addr = %#x  chan = %#x.\n",
              addr, chan );
            return -EINVAL;
           } 
           SetGainQueue( addr, chan, gain );
	} else {
          Chan[minor].gain = (BYTE)arg;
	}
        break;
      case ADC_GET_GAINS:
          put_fs_long((long)Chan[minor].gain, (long*) arg);
          break;
      case ADC_UPDATE_QUEUE:
          if (arg == TRUE ) {
              Chan[minor].update_queue = TRUE;
	  } else {
              Chan[minor].update_queue = FALSE;
          }
          break;
      case ADC_SET_PACER_FREQ:
        if ( arg > MAX_FREQ ) {
            printk("ioctl:  Can not set frequency %ld greater than %d.\n", arg, MAX_FREQ );
            return -1;
        } else {
            BoardData.freq = (LONG)arg;
            SetPacerFreq(&BoardData);
            LoadPacer(&BoardData);               /* load the board frequency now */
        }
        break;
      case ADC_GET_PACER_FREQ:
        put_fs_long(BoardData.freq, (long*) arg);
        break;
      case ADC_STOP_PACER:
        StopPacer();
        break;
      case ADC_CTR0:
        BoardData.statusReg = inb(STATUS_REG);
        if ( arg == 1 ) {
            BoardData.statusReg |=  CTR0;
        } else {
            BoardData.statusReg &= ~(CTR0);
        }
        outb_p(BoardData.statusReg, STATUS_REG);  
        #ifdef DEBUG
          printk("ioctl(ADC_CRT0): STATUS_REG = %#x\n", BoardData.statusReg);
        #endif
        break;
      case ADC_COUNTER0:
        LoadCounter0( (LONG) arg );
        break;
      case ADC_DIO_PRESET:
        if ( arg == -1 ) {
          Chan[minor].adc_dio_pretrig = FALSE;
        } else {
            Chan[minor].adc_dio_pretrig = TRUE;
            Chan[minor].dio_value = (BYTE) (arg & 0xf);
        }

        #ifdef DEBUG
            printk("ioctl(ADC_DIO_PRESET):  value = %#x\n", Chan[minor].dio_value);
        #endif

        break;
      case ADC_GET_STATUS:
        /* return status register */
        put_fs_long( inb_p(STATUS_REG), (long*) arg);
        break;
      case ADC_GET_DIO_VALUE:
          put_fs_long(inb_p(DIO_REG), (long*) arg);
          break;
      case ADC_SET_DIO_VALUE:
          cli();
          bReg = (BYTE) (arg & 0xf);
          outb_p(bReg, DIO_REG);
          sti();
          #ifdef DEBUG
            printk("ADC_DIO set to %#x\n", bReg);
          #endif
          break;
      default:
        return(-EINVAL);
        break;
    } /* end switch */
  }  /* end if */

  if ( minor >= AD_CHANNELS  && minor < AD_CHANNELS + DIO_PORTS ) {
    port = minor - AD_CHANNELS;
    switch (cmd) {
      case DIO_SET_MODE:
        arg &= 0x3;
        switch (port) {
	    case 0:		       /* Port A */
              #ifdef DEBUG
              printk("DIO_SET_MODE for Port A\n");
              #endif
              dio_reg &= 0x9f;
              dio_reg |= (arg << 5);
              outb_p(dio_reg, DIO_CNTRL_REG);
              ChanDIO[port].mode = arg;
              break;

            case 1:			/* Port 1B */
              #ifdef DEBUG
              printk("DIO_SET_MODE for Port 1B\n");
              #endif
              if (arg & 0x2)
	        return(-EINVAL);	       /* Port 1B only has Modes 0 & 1 */
              dio_reg &=  0xfb;
              dio_reg |= (arg << 2);
              outb_p(dio_reg, DIO_CNTRL_REG);
              ChanDIO[port].mode = arg;
              break;

            case 2:			/* Port 1C */
              #ifdef DEBUG
              printk("DIO_SET_MODE for Port 1C\n");
              #endif
              if (arg)
	        return(-EINVAL);	/* Port 1C only has Mode 0 I/O */
              ChanDIO[port].mode = arg;
              break;

            default:
              #ifdef DEBUG
              printk("DIO_SET_MODE for Invalid Port\n");
              #endif
              return(-EINVAL);	/* Wrong Port Number */
              break;
        }
        break;
    
      case DIO_SET_DIRECTION:

        #ifdef DEBUG
        printk("DIO_SET_DIRECTION: minor = %d, arg = %d\n", (int) minor, (int) arg );
        #endif

        switch (port) {
          case 0:			/* Port A */
            #ifdef DEBUG
            printk("DIO_SET_DIRECTION for Port A\n");
            #endif
            arg &= 0x1;
            dio_reg &=  0xef;
            dio_reg |= (arg << 4);
            outb_p(dio_reg, DIO_CNTRL_REG);
            break;

          case 1:			/* Port B */
            #ifdef DEBUG
            printk("DIO_SET_DIRECTION for Port B\n");
            #endif
            arg &= 0x1;
            dio_reg &=  0xfd;
            dio_reg |= (arg << 1);
            outb_p(dio_reg, DIO_CNTRL_REG);
            break;

          case 2:			/* Port C */
            #ifdef DEBUG
            printk("DIO_SET_DIRECTION for Port C\n");
            #endif
            switch ( arg ) {
              case PORT_INPUT:         /* CU = IN  CL = IN */
                  dio_reg |= 0x9;
                  break;
              case PORT_OUTPUT:        /* CU = OUT  CL = OUT */
                  dio_reg &= ~(0x9);
                  break;
              case LOW_PORT_INPUT:     /* CL = IN */
                  dio_reg |= 0x1;
                  break;
              case LOW_PORT_OUTPUT:    /* CL = OUT */
                  dio_reg &= ~(0x1);
                  break;
              case HIGH_PORT_INPUT:    /* CU = IN */
                  dio_reg |= 0x8;
                  break;
              case HIGH_PORT_OUTPUT:   /* CU = OUT */
                  dio_reg &= ~(0x8);
                  break;
              default:
                  return(-EINVAL);
                  break;
            }
            outb_p(dio_reg, DIO_CNTRL_REG);
            break;

          default:
            #ifdef DEBUG
            printk("DIO_SET_DIRECTION for Invalid Port\n");
            #endif
            return(-EINVAL);
            break;
        }
        break;
    
      default:
        return(-EINVAL);
        break;
    }  /* end switch */
  }    /* end if */

  return 0;
}

/***************************************************************************
 *
 * Block copy of data samples from data register to kernel memory.
 *
 ***************************************************************************/

inline char *RegCopy(char *dest, const int reg, int n)
{
__asm__ __volatile__(
	"cld\n\t"
        "rep\n\t"
        "insw"
	: 
	:"D" (dest), "d" (reg),"c" (n)
	:"di","dx","cx");
return dest;
}

/***************************************************************************
 *
 *  Alarm handler to handle situations where an interrupt is missed.
 *
 ***************************************************************************/

static void adc0_TimerHandler(LONG unused)
{
  BYTE bReg;

  /* 
     we should only be here if an interrupt occured while 
     we were reading half the FIFO 
  */

  #ifdef DEBUG
      printk("TimerHandler: WordsToRead = %d\n", WordsToRead);
  #endif

  cli();
 /* turn off interrupts, and trigger control.*/
  bReg = inb(CNTRL_REG) & ~(INTEN | S1 | S0);
  outb_p(bReg, CNTRL_REG);	     
  outb(LSBFIRST, COUNTERB_CONTROL);    /* Set TOTALBAR to 1 */
  outb(0x00, CLR_INTERRUPT);           /* Clear IRQDATA bit  */
  switch (CurrChan->mode) {
    case ADC_EXTERNAL_PACER:
      RegCopy((char *) KernBuffPtr, base_reg, WordsToRead+PACKETSIZE );
      break;
    case ADC_PACER_CLOCK:
    case ADC_PACER_EXTERN_TRIG:
      RegCopy((char *) KernBuffPtr, base_reg, WordsToRead );
      break;
  }
  WordsToRead = 0;
  wake_up_interruptible(&adc0_wait);  
  sti();
}

/***************************************************************************
 *
 * Interrupt handler used to service interrupt read().
 *
 ***************************************************************************/

static void adc0_ReadInterrupt( int irq )
{
  BYTE bReg;

  #ifdef DEBUG
      printk("Entering adc0_ReadInterrupt(). irq = %d mode = %d  count = %d\n", 
              irq, CurrChan->mode, WordsToRead );
  #endif

  cli();
  
  bReg = inb(STATUS_REG) & IRQDATA;
  if ( !bReg ) { /* Should not happen.  Interrupt occured from somewhere else */
    sti();
    printk("adc0_ReadInterrupt: Sperious Interrupt!!!.\n");
    return;
  }

  if ( WordsToRead == 0 ) {
    sti();
    return;
  }
  do {
    switch (CurrChan->mode) {
      case ADC_EXTERNAL_PACER:
        if ( WordsToRead > 2*PACKETSIZE ) {
            RegCopy((char *)KernBuffPtr, base_reg, PACKETSIZE);
            outb(0x1, CLR_INTERRUPT);              /* Clear IRQDATA bit  */
            WordsToRead -= PACKETSIZE;
            KernBuffPtr += PACKETSIZE;
 	} else if ( WordsToRead <= 2*PACKETSIZE && WordsToRead > PACKETSIZE ) {
            WordsToRead -= PACKETSIZE;
            TimerList.expires = jiffies + HZ*60;
            add_timer(&TimerList); 
 	} else {  /* We are done */
            /* turn off interrupts, and trigger control */
            bReg = inb(CNTRL_REG) & ~(INTEN | S1 | S0);
            outb_p(bReg, CNTRL_REG);
            del_timer(&TimerList);
            if (CurrChan->count > PACKETSIZE) {
              RegCopy((char *) KernBuffPtr, base_reg, WordsToRead+PACKETSIZE );
	    } else {
              RegCopy((char *) KernBuffPtr, base_reg, WordsToRead );
	    }
            outb(0x1, CLR_INTERRUPT);                 /* Clear IRQDATA bit  */
            WordsToRead = 0;
            outb(LSBFIRST, COUNTERB_CONTROL);         /* Set TOTALBAR to 1 */
            wake_up_interruptible(&adc0_wait);  
	}
        break;

      case ADC_PACER_CLOCK:
      case ADC_PACER_EXTERN_TRIG:
        if ( WordsToRead > PACKETSIZE ) {
	    RegCopy((char *)KernBuffPtr, base_reg, PACKETSIZE);
            outb(0x1, CLR_INTERRUPT);              /* Clear IRQDATA bit  */
            WordsToRead -= PACKETSIZE;
            KernBuffPtr += PACKETSIZE;
            if ( WordsToRead < PACKETSIZE ) {
                TimerList.expires = jiffies + WordsToRead*HZ/BoardData.freq + 2;
		add_timer(&TimerList);  
            }
         } else {  /* We are done */
            /* turn off interrupts, and trigger control.*/
            bReg = inb(CNTRL_REG) & ~(INTEN | S1 | S0);
            outb(bReg, CNTRL_REG);
            del_timer(&TimerList);
            RegCopy((char *) KernBuffPtr, base_reg, WordsToRead );
            outb(0x1, CLR_INTERRUPT);              /* Clear IRQDATA bit  */
            WordsToRead = 0;
            wake_up_interruptible(&adc0_wait);  
        }
        break;
      default:
        break;
    } 

    /* Check for overflows */
    if ( (bReg = (inb_p(STATUS_REG) & OVRUN)) ) {
      sti();
      printk("adc0_ReadInterrupt: FIFO Buffer overflow or DT-Connect handshake failed.\n");
      outb(bReg & ~(DTEN), STATUS_REG);    /* clear OVRUN bit */
      cli();
    }
  } while (inb(STATUS_REG) & IRQDATA);   /* make sure there is no more data */

  sti();
  return;
}

/***************************************************************************
 *
 * Handles software/pacer triggered read().
 *
 * Bang! Bang! Method: 
 *    
 *    o Force interrupt by forcing conversion.
 *    o Get one sample at a time and put it in the kernel buffer(kernBuff).
 *    o Get chan->count samples
 *
 ***************************************************************************/

static int SoftRead(WORD *kernBuff, ADC_ChanRec *chanRec)
{
  BYTE bReg;

  #ifdef DEBUG
      printk("Entering SoftRead().\n");
  #endif

  BoardData.statusReg = inb(STATUS_REG) & ~TRIG0 ; 
  outb(BoardData.statusReg, STATUS_REG);

  /* Disable interrrupts */
  bReg = inb(CNTRL_REG) & ~(INTEN | S1 | S0);
  #ifdef DEBUG
    printk("SoftRead: CNTRL_REG = %#x   STATUS_REG = %#x, base_reg = %#x.\n",
            bReg, BoardData.statusReg, base_reg);
  #endif

  /* 
     magic write from Jerry Reilly Jerry@computerboards.com  Disables
     total_count so IRQDATA will be set on conversion. This is
     not documented and not obvious.
  */

  outb(LSBFIRST, COUNTERB_CONTROL);

  WordsToRead = chanRec->count;
  KernBuffPtr = KernBuff;

  while ( WordsToRead ) {
    outb(0xff, base_reg);                 /* Force first conversion */
    /* 
       Poll for conversion.  Normally, I would use a interrupt
       handler, but the DAS16-M1 board is too fast, and we get
       a race condition.  Much better to poll for a single conversion
       here.
   */
    do {
      bReg = inb(STATUS_REG) & IRQDATA;
    } while (!bReg);

    *KernBuffPtr++ = (WORD) inw(DATA_AND_CHNLS);  /* Load into buffer */
    outb(0x00, CLR_INTERRUPT);                    /* Clear IRQDATA bit  */
    WordsToRead--;
  }

  return 0;
}

/***************************************************************************
 *
 * Set voltage gains.
 *
 ***************************************************************************/

static void SetGainQueue( BYTE addr, BYTE chan, BYTE gain )
{
  BYTE mask;

  mask = (gain << 4);
  mask |= chan;

  #ifdef DEBUG
      printk("SetGainQueue(): addr = %#x, chan = %#x, gain = %#x,  mask = %#x\n",
              addr, chan, gain, mask);
  #endif

 /* 
    If more than one channel is in the channel gain, the following rules apply:

   1. There must be an even number of entries in the queue.
   2. Even channels must be at even queue addresses.
   3. Add channels must be at odd queue addresses.
 */

  outb(addr, GAIN_Q_ADDR_REG);
  outb(mask, GAIN_Q_DATA_REG);
}

/****************************************************************************
 *   Set Pacer Clock Frequency
 * 
 *   Description: 
 *       Set the counters so that the pacer clock runs at the
 *       desired frequency.  The frequency is generated by dividing
 *       down a 10 MHz clock, so all frequencies can not be generated.
 *       This routine calculated the divisor to generate a frequency
 *       as near as possible to the requested one.  It then calculates
 *       the real frequency and returns it .
 ****************************************************************************/

static int SetPacerFreq(ADC_BoardRec *board)
{
  short unsigned ctr1, ctr2;
  LONG product, error, error2;

  /* divide 10Mhz by frequency */
  product =  (long) 10000000 / board->freq;

  /* check for rounding error */
  error = abs(10000000 - product*board->freq);
  error2 = abs(10000000 - (product+1)*board->freq);
  if ( error2 < error ) product++;

  /* Now the job is to find two 16 bit numbers, that when multiplied
     together are approximately equal to product.  Start by setting
     one of them, ctr1 to 2 (minimum settable value) and increment until
     the error is minimized and ctr2 is less than 32768.

     NOTE: In Mode 2, a value of 1 is illegal! Therefore, crt1 and crt2
     can never be 1.

  */

  ctr1 = product / 32768;
  if ( ctr1 < 2 ) ctr1 = 2;
  ctr2 = product / ctr1;
  error = abs(product - (long) ctr2 * (long) ctr1);

  while ( error && ctr1 < 32768  && ctr2 > 1 ) {
       ctr1++;
       ctr2 = product / ctr1;
       error = abs(product - (long) ctr2 * (long) ctr1);
  }

  /* the frequency is prime, add 1 to it */
  if ( error ) {
      product++;
      ctr1 = product / 32768;
      if ( ctr1 < 2 ) ctr1 = 2;
      ctr2 = product / ctr1;
      error = abs(product - (long) ctr2 * (long) ctr1);

      while ( error  && ctr1 < 32768 && ctr2 > 1) {
          ctr1++;
          ctr2 = product / ctr1;
          error = abs(product - (long) ctr2 * (long) ctr1);
      }
  }

  /* we can't have ctr2 equal to 1, or system hangs */
  if ( ctr2 == 1 ) {
      ctr2++;
      ctr1 /= 2;
  }

  board->ctr1 = ctr1;
  board->ctr2 = ctr2;
  board->freq = 10000000/((long)ctr1*(long)ctr2);

  #ifdef DEBUG
      printk("SetPacerFreq: Pacer Frequecy  set to %ld\n", BoardData.freq);
  #endif

  return 0;
}

/***************************************************************************
 *
 * Load two part frequency to pacer counter chip.
 *
 ***************************************************************************/

static void LoadPacer(ADC_BoardRec *board)
{
  BYTE mask;
  BYTE bData;

  /* Write the values of ctr1 and ctr2 into counter A1 and A2 */

  #ifdef DEBUG
      printk("LoadPacer: load values: ctr1 %#x ctr2 %#x\n", board->ctr1, board->ctr2);
  #endif

  mask = C2+MODE2+LSBFIRST;
  outb_p(mask, COUNTERA_CONTROL); 

  bData = (BYTE) (board->ctr2 & 0xff);
  outb_p(bData, COUNTERA_2_DATA); 

  bData = (BYTE) (board->ctr2 >> 8);
  outb_p(bData, COUNTERA_2_DATA); 

  mask = C1+MODE2+LSBFIRST;
  outb_p(mask, COUNTERA_CONTROL); 

  bData = (board->ctr1 & 0xff);
  outb_p(bData, COUNTERA_1_DATA); 

  bData = (board->ctr1 >> 8) & 0xff;
  outb_p(bData, COUNTERA_1_DATA); 

}

/***************************************************************************
 *
 * Load value into Counter 0  XXXX    Mode    MSB     LSB
 *                            Byte 3  Byte 2  Byte 1  Byte 0
 *
 ***************************************************************************/

static void LoadCounter0( LONG value )
{
  BYTE mask;
  BYTE bData;

  /* Write the value into Counter 0 Mode 2 */

  #ifdef DEBUG
      printk("LoadCounter0: load value %#lx into Counter 0.\n", value);
  #endif

  /* the mode is in the thrid byte */
  mask = 0xff & (value >> 16);
  mask += C0+LSBFIRST;
  outb_p(mask, COUNTERA_CONTROL); 

  if ( value & 0xff000000 ) {       /* load control word only */
      return;
  } else {
      /* LSB in the first byte of value */
      bData = (BYTE) (value & 0xff);
      outb_p(bData, COUNTERA_0_DATA); 

      /* MSB in the second byte of value */
      bData = (BYTE) ((value >> 8) & 0xff);
      outb_p(bData, COUNTERA_0_DATA); 
  }
}

/***************************************************************************
 *
 *  Load the toatl count.  Initialize Total Counter on DAS 16/M1
 *
 * The lower counter operates in Mode 2, if loaded with 0, it will roll over
 * every 65536 counts. Mode 2 output is low when counter = 1, therefore Mode 2
 * does not allow a '1' to be loaded.  Mode 2 is repetitive, where the output
 * will pulse low on every TC. The counter will go from 1 to its reload value,
 * which can be 0 (for 65536) down to 2.
 *
 * The upper counter operates  in Mode 0, which starts out low. The output will
 * go high when the count reaches 0 (not 1 as in Mode 2). The output will then
 * stay high, Mode 0 is not repetitive like Mode 2.
 *
 * The setup of the two counters will be :
 *
 * Load upper counter (Counter 0) with the value from the upper word of the
 * total count, which represents how many multiples of 64k will be counted.
 *
 * Force a clock to upper counter to load it:
 * The lower counter is manipulated to force a clock cycle to the upper counter;
 * this loads the upper counter but does not decrement it, thus loading it with
 * a zero value in Mode 0 will require two clocks to bring the upper counter
 * to Terminal count, the first to load as just described, the second will be
 * the rollover action of the lower count.
 *
 * Load lower counter with residual:
 * The lower counter will first be loaded with the residual (the actual value
 * of the low word of the total count) into the lower counter. The lower
 * counter is then clocked by causing an adc conversion, which causes the
 * residual to be loaded into the lower counter, but does not decrement
 * the counter. If the lower word of the total count is 1 then the lower
 * counter will be loaded with 2 and two adc conversions will be applied, note
 * that when the second clock is applied, the lower counter's output will go low,
 * as the count value will be 1.
 *
 * Please note that since an adc conversion is being done that the channel
 * gain queue must be reestablished after the counters have been setup.
 *
 * Load lower counter with 0 to rollover at 64k:
 * If the total count is greater than 64k (65536) the lower counter is then
 * loaded with a value of 0 (zero), which will be the reload value for the lower
 * counter when it next reaches terminal count. Thus after the initial
 * terminal count (which will occur when the lower counts from the residual
 * down to 1), the lower counter will repetively roll-over every 65536
 * conversions.
 *
 * The output of Ctr1 (lower) is inverted to be the clock for Ctr0 (upper).
 * during setup Ctr1 is forced from one mode to another to change its output
 *
 *
 * Review cascade relationship of Ctr1 (lower) to Ctr0 (upper)
 *              Ctr1 Mode       Ctr1 output     Ctr0 clock input
 *
 *                  0                 low                high
 *                  2                 high               low
 * Review of 8254 usable modes:
 *                       Mode 0                   Mode 2
 *
 *  Best for             Counting down once       Counting down repetively
 *  Running mode         Stops at TC              Reloads repetively
 *  Output Starts        Low                      High
 *  Output at TC         High (and stays there)   Low when count = 1
 *  Output after TC      High (doesn't change)    Back to high
 *  TC value             0                        1
 *  Don't load at        any # ok                 1
 *  Loading 0 means      Out hi after 65536 clks  Reload every 65536 clocks
 *
 * It is most important to understand a proper setup sequence for the 8254
 *           First set Mode in Control Register, this stops the counter
 *           Make sure that clock input is in known state (preferably low)
 *           Next write Load data to Counter,
 *              data gets loaded on the first valid clock 'cycle'
 *              2nd clock cycle yields Count value of (Load-1)
 *              A valid clock is a rising edge followed by a falling edge
 *              all action is relative to the falling edge
 *              A valid clocking of the upper counter (Ctr 0) by manipulating
 *              the lower counter is:
 *
 *        (assumes counter 0 has been loaded with control and count data
 *              Counter 1 to Mode 2        ensure clock starts low
 *              Counter 1 to Mode 0        rising edge of clock
 *              Counter 1 to Mode 2        falling edge of clock
 *
 *              Now Counter 1 can be configured for residual/rollover etc.
 *
 ***************************************************************************/

static void LoadTotalCount( LONG totalCnt )
{
  BYTE bReg;
  WORD wLoCount;
  WORD wHiCount;

  wHiCount = (WORD) (totalCnt >> 16);      /* get upper word */
  wLoCount = (WORD) (totalCnt & 0xffff);   /* get lower word */


  /* load the Total Counter  for the DAS-16/M1 */
  bReg = C0+MODE0+LSBFIRST;              /* CounterB -> mode 0, DON'T load it */
  outb_p(bReg, COUNTERB_CONTROL);        /* Sets output HIGH? */
  if (wLoCount != 0 ) wHiCount++;        /* increment if lower count is ! zero */
  
  /* Set upper word of total count */
  bReg = (BYTE) (wHiCount & 0xff);
  outb_p(bReg, COUNTERB_0_DATA);         
  bReg = (BYTE) (wHiCount >> 8);
  outb_p(bReg, COUNTERB_0_DATA);
  
  bReg = C1+MODE2+LSBFIRST;	         /* Ctr1 (lower) in Mode 2 (TC = 1) */
  outb_p(bReg, COUNTERB_CONTROL);        /* Makes Counter 0 clock low */
  
  bReg = C1+MODE0+LSBFIRST;              /* Make Ctr 0 clock high */
  outb_p(bReg, COUNTERB_CONTROL);        /* rising edge of clock */

  bReg = C1+MODE2+LSBFIRST;	         /* Make Ctr 0 clock low */
  outb_p(bReg, COUNTERB_CONTROL);        /* Completes 1st clk cycle, loads Ctr 0 */

  if ( wLoCount == 1 ) {
    wLoCount++;			         /* wLoCount can not be 1 */
    bReg = (BYTE) (wLoCount & 0xff);     /* load the lower word */
    outb_p(bReg, COUNTERB_1_DATA);
    outb_p(0x1, base_reg);	         /* Force conversion to get count = 1 */
  } else {
    bReg = (BYTE) (wLoCount & 0xff);     /* load the lower word */
    outb_p(bReg, COUNTERB_1_DATA);
    bReg = (BYTE) (wLoCount >> 8);
    outb_p(bReg, COUNTERB_1_DATA);
  }
  outb_p(0x1, base_reg);	         /* Force a conversion */

  /* Load the rollover mode for totals greater than 64k */
  bReg = 0x0;
  outb_p(bReg, COUNTERB_1_DATA);         /* rollover mode ?? */
  outb_p(bReg, COUNTERB_1_DATA);         /* rollover mode ?? */

  bReg = C1+MODE0+LSBFIRST;	         /* Let LSB ctr to mode 0 so that */
  outb_p(bReg, COUNTERB_CONTROL);        /* output does not toggle. */
}


/***************************************************************************
 *
 * Flush the FIFO
 *
 ***************************************************************************/

static void ResetFIFO()
{
  BYTE bReg;

/* Read the gain queue and rewrite it */
  bReg = inb(GAIN_Q_ADDR_REG);
  outb(bReg, GAIN_Q_ADDR_REG);
  bReg = inb(GAIN_Q_DATA_REG);
  outb(bReg, GAIN_Q_DATA_REG);
}


/***************************************************************************
 *
 * Turn off pacer timer chip.
 *
 ***************************************************************************/

static void StopPacer()
{
  BYTE mask;

  mask = C2+MODE2+LSBFIRST;
  outb_p(mask, COUNTERA_CONTROL);
  mask = C1+MODE2+LSBFIRST;
  outb_p(mask, COUNTERA_CONTROL);
}

/***************************************************************************
 *
 * Handles 
 *    o pacer/counter triggered read() with software start.
 *
 * Pacer Method: 
 *
 *    o Pacer counter controls frequency of sample conversions
 *    o Total counter controls number of samples taken
 *    o Interrupts occur every 512(PACKETSIZE) samples and at last sample
 *    o Get chan->count samples
 *    o Samples are saved in kernel buffer(kernBuff) in readInterrupt()
 *
 ***************************************************************************/

static int PacerRead(WORD *kernBuff, ADC_ChanRec *chanRec)
{
  BYTE bReg = 0x0;
  LONG sampleCnt, remainderCnt;

  #ifdef DEBUG
      printk("Entering PacerRead().\n");
  #endif

  cli();

  /* Prepare global data for parameterless interrupt handler */

  CurrChan = chanRec;                        /* pass chanel number to global */
  KernBuffPtr = KernBuff;                    /* same with KernBuff           */
  WordsToRead = chanRec->count;

  /* 
    We can get into a race condition here.  After an interrupt at FIFO
    half full, we need to make sure we have enough time to read the buffer
    (PACKETSIZE) before the next interrupt.  I am assuming a time of 640 us.
    If there is not enough time, it is better to sample another group and
    only read the nubmer we need 
  */

  /* Need to read at least sampleCnt between interrupts */
  remainderCnt = WordsToRead%PACKETSIZE;
  sampleCnt = BoardData.freq/1562;
  if ( sampleCnt > PACKETSIZE ) sampleCnt = PACKETSIZE;

  if ( sampleCnt > remainderCnt && WordsToRead > PACKETSIZE && remainderCnt ) {
      sampleCnt = WordsToRead - remainderCnt + sampleCnt;
  } else {
      sampleCnt = WordsToRead;
  }

  LoadTotalCount(sampleCnt);
  ResetFIFO();

  #ifdef DEBUG
    sti();
    BoardData.statusReg = inb(STATUS_REG);
    bReg = inb(CNTRL_REG);
    printk("PacerRead: CNTRL_REG = %#x   STATUS_REG = %#x, base_reg = %#x  count = %ld.\n",
            bReg, BoardData.statusReg, base_reg, sampleCnt);
    cli();
  #endif

  outb(0x00, CLR_INTERRUPT);                  /* Clear IRQDATA bit  */
  bReg = (BYTE)inb(CNTRL_REG) | INTEN;        /* Enable interrupts */
  outb(bReg, CNTRL_REG);

  if ( chanRec->adc_dio_pretrig ) {
      outb_p( chanRec->dio_value, DIO_REG );
  }

  LoadPacer(&BoardData);                      /* Establish sample frequency */
  interruptible_sleep_on(&adc0_wait);         /* Block in wait state */

  if ( WordsToRead != 0 ) {
      printk("Timing error in PacerRead: WordsToRead = %d\n", WordsToRead);
      return -1;
  }
  #ifdef DEBUG
    printk("Leaving PacerRead.\n");
  #endif
  return 0;
}


/***************************************************************************
 *
 * Handles 
 *    o pacer/counter triggered read() with external start.
 *    o external triggered read().
 *
 * Pacer Method: 
 *
 *    o Pacer counter controls frequency of sample conversions
 *    o Total counter controls number of samples taken
 *    o Interrupts occur every 512(PACKETSIZE) samples and at last sample
 *    o Get chan->count samples
 *    o Samples are saved in kernel buffer(kernBuff) in readInterrupt()
 *
 ***************************************************************************/

static int ExternRead(WORD *kernBuff, ADC_ChanRec *chanRec)
{
  BYTE bReg = 0x0;
  WORD wReg = 0x0;

  #ifdef DEBUG
      printk("Entering ExternRead().\n");
  #endif

  cli();

  /* load the Total Counter  for the DAS-16/M1 */
  bReg = C0+MODE0+LSBFIRST;	             /* CounterB -> mode 0, DON'T load it */
  outb_p(bReg, COUNTERB_CONTROL);            /* Sets output HIGH? */

  wReg = (WORD) (chanRec->count >> 16);       /* get upper word of count */
  if (chanRec->count & 0xffff) wReg++;        /* increment if lower count is ! zero */

  bReg = C0+MODE0+LSBFIRST;	             /* Counter -> mode 0, DON'T load it */
  outb_p(bReg, COUNTERB_CONTROL);            /* Sets output HIGH? */
  bReg = (BYTE) (wReg & 0xff);	             /* load the upper word */
  outb_p(bReg, COUNTERB_0_DATA);
  bReg = (BYTE) (wReg >> 8);
  outb_p(bReg, COUNTERB_0_DATA);

  /* Set the trigger counters. Counter 0 is changed to counter 1 */

  bReg = C1+MODE2+LSBFIRST;	             /* Ensure clock starts low */
  outb_p(bReg, COUNTERB_CONTROL);    

  bReg = C1+MODE0+LSBFIRST;	             /* Make C0 clock high */
  outb_p(bReg, COUNTERB_CONTROL); 

  bReg = C1+MODE2+LSBFIRST;	             /* Make C0 clock low */
  outb_p(bReg, COUNTERB_CONTROL);            /* complete 1st clock cycle */

  wReg =  (WORD) (chanRec->count & 0xffff);

  if (wReg == 1 ) {
    wReg++;			             /* wReg can not be 1 */
    bReg = (BYTE) (wReg & 0xff);             /* load the lower word */
    outb_p(bReg, COUNTERB_1_DATA);
    bReg = (BYTE) (wReg >> 8);
    outb_p(bReg, COUNTERB_1_DATA);
    outb_p(0x1, base_reg);	             /* Force conversion to get count = 1 */
  } else {
    bReg = (BYTE) (wReg & 0xff);             /* load the lower word */
    outb_p(bReg, COUNTERB_1_DATA);
    bReg = (BYTE) (wReg >> 8);
    outb_p(bReg, COUNTERB_1_DATA);
  }
  outb_p(0x1, base_reg);	             /* Force a conversion */
  outw_p(0x0, COUNTERB_1_DATA);              /* rollover mode ?? */

  /* Prepare global data for parameterless interrupt handler */

  CurrChan = chanRec;                           /* pass chanel number to global */
  KernBuffPtr = KernBuff;                       /* same with KernBuff           */
  WordsToRead = chanRec->count;

  if ( chanRec->adc_dio_pretrig ) {
      /* enable interrupts before setting dio bits */
      outb_p( chanRec->dio_value, DIO_REG );
  }

  #ifdef DEBUG
    sti();
    BoardData.statusReg = inb(STATUS_REG);
    bReg = inb(CNTRL_REG);
    printk("ExternRead: CNTRL_REG = %#x   STATUS_REG = %#x, base_reg = %#x.\n",
            bReg, BoardData.statusReg, base_reg);
    cli();
  #endif

  outb(0x00, CLR_INTERRUPT);                  /* Clear IRQDATA bit  */
  bReg = (BYTE)inb(CNTRL_REG) | INTEN;        /* Enable interrupts */
  outb(bReg, CNTRL_REG);

  interruptible_sleep_on(&adc0_wait);           /* Block in wait state */

  if ( WordsToRead != 0 ) {
      printk("Timing error in ExternRead: WordsToRead = %d\n", WordsToRead);
      return -1;
  }
  #ifdef DEBUG
    printk("Leaving ExternRead.\n");
  #endif
  return 0;
}
