/*
 * Copyright (C) 1997  Warren Jasper
 * All rights reserved.
 *
 * This software may be freely copied, modified, and redistributed
 * provided that this copyright notice is preserved on all copies.
 *
 * You may not distribute this software, in whole or in part, as part of
 * any commercial product without the express consent of the authors.
 *
 * There is no warranty or other guarantee of fitness of this software
 * for any purpose.  It is provided solely "as is".
 */

/***************************************************************************
 *
 *  adcread.c
 *
 *  This program is used to test the CIO-DAC 16/M1 Analog to Digital
 *  Linux loadable module(das16/M1).
 *
 ***************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <sys/types.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include "adc.h"

/***************************************************************************
 *
 *  Global Data
 *
 ***************************************************************************/

char *DevName = "/dev/das16-M1_adc00";
int  Gain     = UP_10_00V;
int  Mode     = ADC_SOFT_TRIGGER;
int  Count    = 1;
int  NoStop   = 0;
int  Print    = 1;
int  Freq     = 100000;
int  Status;

/***************************************************************************
 *
 *  Show user how the command line is used.
 *
 ***************************************************************************/

void Usage( void )
{
    fprintf(stderr, "\n\
Usage: adcread \'options\'\n\
Options:\n\
              [-dev /dev/adc##]  - Specify device file.\n
              [-range (0-8)]     - Voltage input range (see manual)\n
              [-ct ##]           - Number of samples to read\n
              [-soft]            - Use software triggers\n
              [-extern]          - Use extnernal clock triggers\n
              [-pacer]           - Use pacer clock triggers\n
              [-noprint]         - Don't print samples\n
              [-nostop]          - Sample forever\n
              [-freq ##]         - Sample Frequency used with pacer triggers\n
\n");
    exit(1);
}

/***************************************************************************
 *
 *  Process command line switches
 *
 ***************************************************************************/

void DoCommandLine(int argc, char **argv)
{
  int i = 1;

  while (i < argc) {
    if (strcmp(argv[i], "-dev") == 0) {
      i++;
      if (i == argc)
        Usage();
      else 
        DevName = argv[i];
    } else if (strcmp(argv[i], "-range") == 0) {
      i++;
      if (i == argc)
        Usage();
      else 
        Gain = atoi(argv[i]);
    } else if (strcmp(argv[i], "-ct") == 0) {
      i++;
      if (i == argc)
        Usage();
      else 
        Count = atoi(argv[i]);
    } else if (strcmp(argv[i], "-soft") == 0) {
      Mode = ADC_SOFT_TRIGGER;
    } else if (strcmp(argv[i], "-pacer") == 0) {
      Mode = ADC_PACER_CLOCK;
    } else if (strcmp(argv[i], "-extern") == 0) {
      Mode = ADC_EXTERNAL_PACER;
    } else if (strcmp(argv[i], "-noprint") == 0) {
      Print = 0;
    } else if (strcmp(argv[i], "-nostop") == 0) {
      NoStop = 1;
    } else if (strcmp(argv[i], "-freq") == 0) {
      i++;
      if (i == argc)
        Usage();
      else 
        Freq = atoi(argv[i]);
    } else {
      Usage();
    }
    i++;
  }
}

/***************************************************************************
 *
 *  Main
 *
 ***************************************************************************/

int main(int argc, char **argv)
{
  int   i;
  int	fd;
  WORD  data[20000];
  int   bytesRead;
  BYTE  toggle = 0;
  double mean, sd;

  DoCommandLine(argc, argv);
	
  /* Open the device file for the counter */

  printf("Running in mode %d\n", Mode);

  if ((fd = open(DevName, Mode )) < 0) {
    perror(DevName);
    printf("error opening device %s\n", DevName);
    exit(2);
  }


  /* Set voltage input range */

  printf("Setting voltage input range to %d. Look at manual.\n", Gain);
  ioctl(fd, ADC_SET_GAINS, Gain);

  /* program counter 0 */
  /*
  ioctl(fd, ADC_COUNTER0, CTR0_MODE2 | 10) ;
  ioctl(fd, ADC_CTR0, 0x1);
  */
 
  /* Set sampling frequency */
  printf("Requested frequency at %d Hz.\n", Freq);
  ioctl(fd, ADC_SET_PACER_FREQ, Freq);
  ioctl(fd, ADC_GET_PACER_FREQ, &Freq);
  printf("Actual frequency set to %d Hz.\n", Freq);

  /* Get status */
  ioctl(fd, ADC_GET_STATUS, &Status);
  printf("Contents of status register: %#x\n", Status);

  printf("sleep before read\n\n");
  sleep(2);
  
  do {
    toggle ^= 0x2;
    /*    ioctl(fd, ADC_SET_DIO_VALUE, toggle); */

    bytesRead = read(fd, data, Count);
    if (bytesRead != Count) {
      printf("main() in adcread.c: ERROR on read()\n");
      exit(1);
    }

    if (Print) {
      for (i = 0; i < Count; i++){ 
	  printf("Sample %d = %d\n", i, data[i]); 
      }
    }
  } while (NoStop);

  printf("%d samples read\n", bytesRead);
  printf("\n*******  Statistics ******\n");
  mean = 0.0;
  for ( i = 0; i < Count; i++ ) {
      mean += data[i];
  }
  mean /= Count;
  sd = 0.0;
  for ( i = 0; i < Count; i++ ) {
    sd += (data[i] - mean)*(data[i] - mean);
  }
  sd = sqrt( sd / (Count - 1));
  printf("Mean = %f,  Standard deviation = %f.\n\n", mean, sd);

  close(fd);
  return 0;
}
	
	


